@@ -37,7 +37,11 @@ struct connman_device {
 	enum connman_device_type type;
 	enum connman_device_mode mode;
 	connman_bool_t secondary;
+	connman_bool_t offlinemode;
+	connman_bool_t blocked;
 	connman_bool_t powered;
+	connman_bool_t powered_pending;
+	connman_bool_t powered_persistent;
 	connman_bool_t carrier;
 	connman_bool_t scanning;
 	connman_bool_t disconnected;
@@ -48,6 +52,7 @@ struct connman_device {
 	char *interface;
 	char *control;
 	char *ident;
+	int phyindex;
 	unsigned int connections;
 	guint scan_timeout;
 	struct connman_ipconfig *ipconfig;
@@ -82,6 +87,25 @@ static gboolean device_scan_trigger(gpointer user_data)
 	return TRUE;
 }
 
+static void clear_scan_trigger(struct connman_device *device)
+{
+	if (device->scan_timeout > 0) {
+		g_source_remove(device->scan_timeout);
+		device->scan_timeout = 0;
+	}
+}
+
+static void reset_scan_trigger(struct connman_device *device)
+{
+	clear_scan_trigger(device);
+
+	if (device->scan_interval > 0) {
+		guint interval = device->scan_interval;
+		device->scan_timeout = g_timeout_add_seconds(interval,
+					device_scan_trigger, device);
+	}
+}
+
 static const char *type2description(enum connman_device_type type)
 {
 	switch (type) {
@@ -136,6 +160,34 @@ static const char *type2string(enum connman_device_type type)
 	return NULL;
 }
 
+enum connman_service_type __connman_device_get_service_type(struct connman_device *device)
+{
+	enum connman_device_type type = connman_device_get_type(device);
+
+	switch (type) {
+	case CONNMAN_DEVICE_TYPE_UNKNOWN:
+	case CONNMAN_DEVICE_TYPE_VENDOR:
+	case CONNMAN_DEVICE_TYPE_GPS:
+	case CONNMAN_DEVICE_TYPE_NOZOMI:
+	case CONNMAN_DEVICE_TYPE_HUAWEI:
+	case CONNMAN_DEVICE_TYPE_NOVATEL:
+		break;
+	case CONNMAN_DEVICE_TYPE_ETHERNET:
+		return CONNMAN_SERVICE_TYPE_ETHERNET;
+	case CONNMAN_DEVICE_TYPE_WIFI:
+		return CONNMAN_SERVICE_TYPE_WIFI;
+	case CONNMAN_DEVICE_TYPE_WIMAX:
+		return CONNMAN_SERVICE_TYPE_WIMAX;
+	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
+		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+	case CONNMAN_DEVICE_TYPE_MBM:
+	case CONNMAN_DEVICE_TYPE_HSO:
+		return CONNMAN_SERVICE_TYPE_CELLULAR;
+	}
+
+	return CONNMAN_SERVICE_TYPE_UNKNOWN;
+}
+
 static int set_connected(struct connman_device *device,
 						connman_bool_t connected)
 {
@@ -186,6 +238,18 @@ static int set_connected(struct connman_device *device,
 					CONNMAN_SERVICE_STATE_IDLE);
 	}
 
+	if (connected == TRUE) {
+		enum connman_service_type type;
+
+		type = __connman_device_get_service_type(device);
+		__connman_notifier_connect(type);
+	} else {
+		enum connman_service_type type;
+
+		type = __connman_device_get_service_type(device);
+		__connman_notifier_disconnect(type);
+	}
+
 	return 0;
 }
 
@@ -233,30 +297,39 @@ static int powered_changed(struct connman_device *device)
 static int set_powered(struct connman_device *device, connman_bool_t powered)
 {
 	struct connman_device_driver *driver = device->driver;
+	enum connman_service_type type;
 	int err;
 
 	DBG("device %p powered %d", device, powered);
 
-	if (device->powered == powered)
+	if (device->powered_pending == powered)
 		return -EALREADY;
 
+	device->powered_pending = powered;
+
 	if (!driver)
 		return -EINVAL;
 
+	type = __connman_device_get_service_type(device);
+
 	if (powered == TRUE) {
 		if (driver->enable) {
 			err = driver->enable(device);
-			__connman_notifier_device_type_increase(device->type);
+			if (err == 0)
+				__connman_notifier_enable(type);
 		} else
 			err = -EINVAL;
 	} else {
+		clear_scan_trigger(device);
+
 		g_hash_table_remove_all(device->networks);
 
 		set_carrier(device, FALSE);
 
 		if (driver->disable) {
 			err = driver->disable(device);
-			__connman_notifier_device_type_decrease(device->type);
+			if (err == 0)
+				__connman_notifier_disable(type);
 		} else
 			err = -EINVAL;
 	}
@@ -425,6 +498,10 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&value, &powered);
 
+		device->powered_persistent = powered;
+
+		__connman_storage_save_device(device);
+
 		if (device->powered == powered)
 			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 
@@ -460,17 +537,12 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&value, &interval);
 
-		device->scan_interval = interval;
+		if (device->scan_interval != interval) {
+			device->scan_interval = interval;
 
-		if (device->scan_timeout > 0) {
-			g_source_remove(device->scan_timeout);
-			device->scan_timeout = 0;
-		}
+			__connman_storage_save_device(device);
 
-		if (device->scan_interval > 0) {
-			guint interval = device->scan_interval;
-			device->scan_timeout = g_timeout_add_seconds(interval,
-						device_scan_trigger, device);
+			reset_scan_trigger(device);
 		}
 	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
 		int err;
@@ -491,213 +563,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 	} else
 		return __connman_error_invalid_property(msg);
 
-	__connman_storage_save_device(device);
-
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static void convert_name(const char *ssid, char *name,
-						unsigned int ssid_len)
-{
-	unsigned int i;
-
-	for (i = 0; i < ssid_len; i++) {
-		if (g_ascii_isprint(ssid[i]))
-			name[i] = ssid[i];
-		else
-			name[i] = ' ';
-	}
-}
-
-static struct connman_network *find_network(struct connman_device *device,
-					const void *ssid, const char *address,
-					const char *security, const char *mode)
-{
-	GHashTableIter network_iter;
-	gpointer key, value;
-
-	if (ssid == NULL)
-		return NULL;
-
-	g_hash_table_iter_init(&network_iter, device->networks);
-	while (g_hash_table_iter_next(&network_iter, &key, &value) == TRUE) {
-		const void *tmp_ssid;
-		const char *tmp_security, *tmp_mode, *tmp_address;
-		unsigned int tmp_ssid_size;
-
-		tmp_ssid = connman_network_get_blob(value, "WiFi.SSID",
-							&tmp_ssid_size);
-		tmp_security = connman_network_get_string(value,
-							"WiFi.Security");
-		tmp_mode = connman_network_get_string(value, "WiFi.Mode");
-		tmp_address = connman_network_get_string(value, "Address");
-
-		if (tmp_ssid_size < 1)
-			continue;
-
-		if (tmp_ssid && memcmp(ssid, tmp_ssid, tmp_ssid_size))
-			continue;
-
-		if (security && tmp_security &&
-				g_strcmp0(security, tmp_security) != 0)
-			continue;
-
-		if (mode && tmp_mode && g_strcmp0(mode, tmp_mode) != 0)
-			continue;
-
-		if (address && tmp_address &&
-				g_strcmp0(address, tmp_address) != 0)
-			continue;
-
-		return connman_network_ref(value);
-	}
-
-	return NULL;
-}
-
-static DBusMessage *join_network(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	struct connman_device *device = data;
-	struct connman_network *network, *found_network;
-	enum connman_network_type type;
-	unsigned int ssid_size;
-	const char *group, *mode, *security, *address;
-	const void *ssid;
-	DBusMessageIter iter, array;
-	int err, index;
-
-	DBG("conn %p", conn);
-
-	if (__connman_security_check_privilege(msg,
-					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
-		return __connman_error_permission_denied(msg);
-
-	if (!device->driver || !device->driver->join)
-		return __connman_error_not_supported(msg);
-
-	dbus_message_iter_init(msg, &iter);
-	dbus_message_iter_recurse(&iter, &array);
-
-	switch (device->type) {
-	case CONNMAN_DEVICE_TYPE_WIFI:
-		type = CONNMAN_NETWORK_TYPE_WIFI;
-		break;
-	default:
-		return __connman_error_not_supported(msg);
-	}
-
-	network = connman_network_create("00_00_00_00_00_00", type);
-	if (network == NULL)
-		return __connman_error_failed(msg, ENOMEM);
-
-	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_DICT_ENTRY) {
-		DBusMessageIter entry, value;
-		const char *key, *str;
-
-		dbus_message_iter_recurse(&array, &entry);
-		dbus_message_iter_get_basic(&entry, &key);
-
-		dbus_message_iter_next(&entry);
-		dbus_message_iter_recurse(&entry, &value);
-
-		switch (dbus_message_iter_get_arg_type(&value)) {
-		case DBUS_TYPE_STRING:
-			dbus_message_iter_get_basic(&value, &str);
-			if (g_str_equal(key, "WiFi.SSID") == TRUE) {
-				char *name;
-
-				connman_network_set_blob(network, key,
-							str, strlen(str));
-				name = g_try_malloc0(strlen(str) + 1);
-				if (name == NULL)
-					return __connman_error_failed(msg,
-								      -ENOMEM);
-
-				convert_name((char *) str, name, strlen(str));
-				connman_network_set_name(network, name);
-				g_free(name);
-
-			} else
-				connman_network_set_string(network, key, str);
-			break;
-		}
-
-		dbus_message_iter_next(&array);
-	}
-
-	ssid = connman_network_get_blob(network, "WiFi.SSID", &ssid_size);
-	security = connman_network_get_string(network, "WiFi.Security");
-	mode = connman_network_get_string(network, "WiFi.Mode");
-	address = connman_network_get_string(network, "Address");
-
-	found_network = find_network(device, ssid, address, security, mode);
-	if (found_network != NULL) {
-		const char* passphrase;
-
-		passphrase = connman_network_get_string(network,
-							"WiFi.Passphrase");
-		if (passphrase != NULL)
-			connman_network_set_string(found_network,
-							"WiFi.Passphrase",
-								passphrase);
-		connman_network_unref(network);
-		network = found_network;
-
-		err = __connman_network_connect(found_network);
-
-		connman_network_unref(found_network);
-
-		if (err < 0)
-			return __connman_error_failed(msg, -err);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
-
-	group = connman_wifi_build_group_name(ssid, ssid_size, mode, security);
-
-	index = connman_device_get_index(device);
-	connman_network_set_index(network, index);
-
-	connman_network_set_protocol(network, CONNMAN_NETWORK_PROTOCOL_IP);
-
-	if (connman_device_get_disconnected(device) == FALSE)
-		__connman_device_disconnect(device);
-
-	err = device->driver->join(device, network);
-
-	connman_network_unref(network);
-
-	if (err < 0)
-		return __connman_error_failed(msg, -err);
-
-	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-}
-
-static DBusMessage *create_network(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	DBG("conn %p", conn);
-
-	if (__connman_security_check_privilege(msg,
-					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
-		return __connman_error_permission_denied(msg);
-
-	return __connman_error_invalid_arguments(msg);
-}
-
-static DBusMessage *remove_network(DBusConnection *conn,
-					DBusMessage *msg, void *data)
-{
-	DBG("conn %p", conn);
-
-	if (__connman_security_check_privilege(msg,
-					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
-		return __connman_error_permission_denied(msg);
-
-	return __connman_error_invalid_arguments(msg);
-}
-
 static DBusMessage *propose_scan(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -726,9 +594,6 @@ static GDBusMethodTable device_methods[] = {
 	{ "GetProperties", "",      "a{sv}", get_properties },
 	{ "SetProperty",   "sv",    "",      set_property,
 						G_DBUS_METHOD_FLAG_ASYNC },
-	{ "JoinNetwork",   "a{sv}", "",      join_network   },
-	{ "CreateNetwork", "a{sv}", "o",     create_network },
-	{ "RemoveNetwork", "o",     "",      remove_network },
 	{ "ProposeScan",   "",      "",      propose_scan   },
 	{ },
 };
@@ -809,36 +674,6 @@ static void unregister_interface(struct connman_element *element)
 						CONNMAN_DEVICE_INTERFACE);
 }
 
-static void device_enable(struct connman_device *device)
-{
-	DBG("device %p", device);
-
-	if (device->powered == TRUE)
-		return;
-
-	if (device->driver->enable) {
-		if (device->driver->enable(device) == 0)
-			device->powered = TRUE;
-		__connman_notifier_device_type_increase(device->type);
-	}
-}
-
-static void device_disable(struct connman_device *device)
-{
-	DBG("device %p", device);
-
-	if (device->powered == FALSE)
-		return;
-
-	g_hash_table_remove_all(device->networks);
-
-	if (device->driver->disable) {
-		if (device->driver->disable(device) == 0)
-			device->powered = FALSE;
-		__connman_notifier_device_type_decrease(device->type);
-	}
-}
-
 static int setup_device(struct connman_device *device)
 {
 	int err;
@@ -864,7 +699,9 @@ static int setup_device(struct connman_device *device)
 		break;
 	}
 
-	device_enable(device);
+	if (device->offlinemode == FALSE &&
+				device->powered_persistent == TRUE)
+		__connman_device_enable(device);
 
 	return 0;
 }
@@ -896,7 +733,7 @@ static void remove_device(struct connman_device *device)
 {
 	DBG("device %p", device);
 
-	device_disable(device);
+	__connman_device_disable(device);
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
@@ -1070,6 +907,10 @@ struct connman_device *connman_device_create(const char *node,
 	device->mode      = CONNMAN_DEVICE_MODE_UNKNOWN;
 	device->secondary = FALSE;
 
+	device->powered_persistent = TRUE;
+
+	device->phyindex = -1;
+
 	switch (type) {
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
@@ -1195,6 +1036,17 @@ int connman_device_get_index(struct connman_device *device)
 	return device->element.index;
 }
 
+int __connman_device_get_phyindex(struct connman_device *device)
+{
+	return device->phyindex;
+}
+
+void __connman_device_set_phyindex(struct connman_device *device,
+							int phyindex)
+{
+	device->phyindex = phyindex;
+}
+
 /**
  * connman_device_set_interface:
  * @device: device structure
@@ -1305,6 +1157,8 @@ connman_bool_t connman_device_get_secondary(struct connman_device *device)
 int connman_device_set_powered(struct connman_device *device,
 						connman_bool_t powered)
 {
+	enum connman_service_type type;
+
 	DBG("driver %p powered %d", device, powered);
 
 	if (device->timeout > 0) {
@@ -1324,6 +1178,14 @@ int connman_device_set_powered(struct connman_device *device,
 		return -EALREADY;
 
 	device->powered = powered;
+	device->powered_pending = powered;
+
+	type = __connman_device_get_service_type(device);
+
+	if (device->powered == TRUE)
+		__connman_notifier_enable(type);
+	else
+		__connman_notifier_disable(type);
 
 	if (device->registered == FALSE)
 		return 0;
@@ -1333,16 +1195,7 @@ int connman_device_set_powered(struct connman_device *device,
 	if (powered == FALSE)
 		return 0;
 
-	if (device->scan_timeout > 0) {
-		g_source_remove(device->scan_timeout);
-		device->scan_timeout = 0;
-	}
-
-	if (device->scan_interval > 0) {
-		guint interval = device->scan_interval;
-		device->scan_timeout = g_timeout_add_seconds(interval,
-					device_scan_trigger, device);
-	}
+	reset_scan_trigger(device);
 
 	if (device->driver->scan)
 		device->driver->scan(device);
@@ -1350,6 +1203,26 @@ int connman_device_set_powered(struct connman_device *device,
 	return 0;
 }
 
+int __connman_device_set_blocked(struct connman_device *device,
+						connman_bool_t blocked)
+{
+	connman_bool_t powered;
+
+	DBG("device %p blocked %d", device, blocked);
+
+	device->blocked = blocked;
+
+	if (device->offlinemode == TRUE)
+		return 0;
+
+	if (blocked == FALSE)
+		powered = device->powered_persistent;
+	else
+		powered = FALSE;
+
+	return set_powered(device, powered);
+}
+
 /**
  * connman_device_set_carrier:
  * @device: device structure
@@ -1390,6 +1263,89 @@ int __connman_device_scan(struct connman_device *device)
 	return device->driver->scan(device);
 }
 
+int __connman_device_enable(struct connman_device *device)
+{
+	enum connman_service_type type;
+	int err;
+
+	DBG("device %p", device);
+
+	if (!device->driver || !device->driver->enable)
+		return -EOPNOTSUPP;
+
+	if (device->powered_pending == TRUE)
+		return -EALREADY;
+
+	device->powered_pending = TRUE;
+
+	err = device->driver->enable(device);
+	if (err < 0)
+		return err;
+
+	device->powered = TRUE;
+
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_enable(type);
+
+	return 0;
+}
+
+int __connman_device_enable_persistent(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	device->powered_persistent = TRUE;
+
+	__connman_storage_save_device(device);
+
+	return __connman_device_enable(device);
+}
+
+int __connman_device_disable(struct connman_device *device)
+{
+	enum connman_service_type type;
+	int err;
+
+	DBG("device %p", device);
+
+	if (!device->driver || !device->driver->disable)
+		return -EOPNOTSUPP;
+
+	if (device->powered == FALSE)
+		return -ENOLINK;
+
+	if (device->powered_pending == FALSE)
+		return -EALREADY;
+
+	device->powered_pending = FALSE;
+
+	clear_scan_trigger(device);
+
+	g_hash_table_remove_all(device->networks);
+
+	err = device->driver->disable(device);
+	if (err < 0)
+		return err;
+
+	device->powered = FALSE;
+
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_disable(type);
+
+	return 0;
+}
+
+int __connman_device_disable_persistent(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	device->powered_persistent = FALSE;
+
+	__connman_storage_save_device(device);
+
+	return __connman_device_disable(device);
+}
+
 int __connman_device_connect(struct connman_device *device)
 {
 	DBG("device %p", device);
@@ -1417,9 +1373,7 @@ int __connman_device_disconnect(struct connman_device *device)
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		struct connman_network *network = value;
 
-		if (connman_network_get_connected(network) == TRUE ||
-				connman_network_get_associating(network) == TRUE)
-			__connman_network_disconnect(network);
+		__connman_network_disconnect(network);
 	}
 
 	if (device->driver && device->driver->disconnect)
@@ -1453,6 +1407,12 @@ static gboolean remove_unavailable_network(gpointer key, gpointer value,
 	return TRUE;
 }
 
+void __connman_device_cleanup_networks(struct connman_device *device)
+{
+	g_hash_table_foreach_remove(device->networks,
+					remove_unavailable_network, NULL);
+}
+
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1494,24 +1454,15 @@ int connman_device_set_scanning(struct connman_device *device,
 	g_dbus_send_message(connection, signal);
 
 	if (scanning == TRUE) {
-		if (device->scan_timeout > 0) {
-			g_source_remove(device->scan_timeout);
-			device->scan_timeout = 0;
-		}
-
-		if (device->scan_interval > 0) {
-			guint interval = device->scan_interval;
-			device->scan_timeout = g_timeout_add_seconds(interval,
-						device_scan_trigger, device);
-		}
+		reset_scan_trigger(device);
 
 		g_hash_table_foreach(device->networks,
 					mark_network_unavailable, NULL);
+
 		return 0;
 	}
 
-	g_hash_table_foreach_remove(device->networks,
-					remove_unavailable_network, NULL);
+	__connman_device_cleanup_networks(device);
 
 	if (device->connections > 0)
 		return 0;
@@ -1519,6 +1470,8 @@ int connman_device_set_scanning(struct connman_device *device,
 	if (device->disconnected == TRUE)
 		return 0;
 
+	__connman_service_auto_connect();
+
 	return 0;
 }
 
@@ -1649,11 +1602,16 @@ static void set_offlinemode(struct connman_element *element, gpointer user_data)
 	if (device == NULL)
 		return;
 
+	device->offlinemode = offlinemode;
+
 	powered = (offlinemode == TRUE) ? FALSE : TRUE;
 
 	if (device->powered == powered)
 		return;
 
+	if (device->powered_persistent == FALSE)
+		powered = FALSE;
+
 	set_powered(device, powered);
 }
 
@@ -1664,19 +1622,29 @@ int __connman_device_set_offlinemode(connman_bool_t offlinemode)
 	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
 			set_offlinemode, GUINT_TO_POINTER(offlinemode));
 
-	__connman_notifier_offline_mode(offlinemode);
+	__connman_notifier_offlinemode(offlinemode);
 
 	return 0;
 }
 
 void __connman_device_increase_connections(struct connman_device *device)
 {
+	enum connman_service_type type;
+
 	device->connections++;
+
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_connect(type);
 }
 
 void __connman_device_decrease_connections(struct connman_device *device)
 {
+	enum connman_service_type type;
+
 	device->connections--;
+
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_disconnect(type);
 }
 
 /**
@@ -1705,8 +1673,6 @@ int connman_device_add_network(struct connman_device *device,
 
 	__connman_network_set_device(network, device);
 
-	__connman_storage_load_network(network);
-
 	err = connman_element_register((struct connman_element *) network,
 							&device->element);
 	if (err < 0) {
@@ -1786,19 +1752,14 @@ void __connman_device_set_network(struct connman_device *device,
  */
 int connman_device_register(struct connman_device *device)
 {
+	enum connman_service_type type;
+
 	__connman_storage_load_device(device);
 
-	switch (device->mode) {
-	case CONNMAN_DEVICE_MODE_UNKNOWN:
-	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
-		break;
-	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
-	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
-		__connman_storage_init_network(device);
-		break;
-	}
+	device->offlinemode = __connman_manager_get_offlinemode();
 
-	__connman_notifier_device_type_register(device->type);
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_register(type);
 
 	return connman_element_register(&device->element, NULL);
 }
@@ -1811,9 +1772,12 @@ int connman_device_register(struct connman_device *device)
  */
 void connman_device_unregister(struct connman_device *device)
 {
+	enum connman_service_type type;
+
 	__connman_storage_save_device(device);
 
-	__connman_notifier_device_type_unregister(device->type);
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_unregister(type);
 
 	connman_element_unregister(&device->element);
 }
@@ -1910,38 +1874,27 @@ static struct connman_driver device_driver = {
 static int device_load(struct connman_device *device)
 {
 	GKeyFile *keyfile;
-	gchar *pathname, *identifier, *data = NULL;
-	gsize length;
+	GError *error = NULL;
+	gchar *identifier;
+	connman_bool_t powered;
 	int val;
 
 	DBG("device %p", device);
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-					__connman_profile_active_ident());
-	if (pathname == NULL)
-		return -ENOMEM;
-
-	keyfile = g_key_file_new();
-
-	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE) {
-		g_free(pathname);
-		return -ENOENT;
-	}
-
-	g_free(pathname);
-
-	if (g_key_file_load_from_data(keyfile, data, length,
-							0, NULL) == FALSE) {
-		g_free(data);
-		return -EILSEQ;
-	}
-
-	g_free(data);
+	keyfile = __connman_storage_open();
+	if (keyfile == NULL)
+		return 0;
 
 	identifier = g_strdup_printf("device_%s", device->element.name);
 	if (identifier == NULL)
 		goto done;
 
+	powered = g_key_file_get_boolean(keyfile, identifier,
+						"Powered", &error);
+	if (error == NULL)
+		device->powered_persistent = powered;
+	g_clear_error(&error);
+
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
 	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
@@ -1949,51 +1902,39 @@ static int device_load(struct connman_device *device)
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
 		val = g_key_file_get_integer(keyfile, identifier,
-							"ScanInterval", NULL);
-		if (val > 0)
+						"ScanInterval", &error);
+		if (error == NULL && val > 0)
 			device->scan_interval = val;
+		g_clear_error(&error);
 		break;
 	}
 
 done:
-	g_key_file_free(keyfile);
-
 	g_free(identifier);
 
+	__connman_storage_close(keyfile, FALSE);
+
 	return 0;
 }
 
 static int device_save(struct connman_device *device)
 {
 	GKeyFile *keyfile;
-	gchar *pathname, *identifier = NULL, *data = NULL;
-	gsize length;
+	gchar *identifier;
 
 	DBG("device %p", device);
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-					__connman_profile_active_ident());
-	if (pathname == NULL)
-		return -ENOMEM;
+	keyfile = __connman_storage_open();
+	if (keyfile == NULL)
+		return 0;
 
-	keyfile = g_key_file_new();
-
-	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE)
-		goto update;
-
-	if (length > 0) {
-		if (g_key_file_load_from_data(keyfile, data, length,
-							0, NULL) == FALSE)
-			goto done;
-	}
-
-	g_free(data);
-
-update:
 	identifier = g_strdup_printf("device_%s", device->element.name);
 	if (identifier == NULL)
 		goto done;
 
+	g_key_file_set_boolean(keyfile, identifier,
+					"Powered", device->powered_persistent);
+
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
 	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
@@ -2006,18 +1947,10 @@ update:
 		break;
 	}
 
-	data = g_key_file_to_data(keyfile, &length, NULL);
-
-	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
-		connman_error("Failed to store device information");
-
 done:
-	g_free(data);
-
-	g_key_file_free(keyfile);
-
 	g_free(identifier);
-	g_free(pathname);
+
+	__connman_storage_close(keyfile, TRUE);
 
 	return 0;
 }
